/*

  Name: OpenCLBase

  Author: Uldis Locans

  Info: OpenCL base class to handle all the common details associated 
  with kernel launch on OpenCL device

  Date: 2014.09.18

*/

#ifndef H_OPENCL_BASE
#define H_OPENCL_BASE

#include <iostream>
#include <iomanip>
#include <vector>
#include <algorithm>
#include <string.h>
#include <stdio.h>


#ifdef __APPLE__
#include <OpenCL/opencl.h>
#include <OpenCL/cl_ext.h>
#else
#include <CL/cl.h>
#include <CL/cl_ext.h>
#endif

#include "../DKSDefinitions.h"

/* struct for random number state */
typedef struct {
  double s10;
  double s11;
  double s12;
  double s20;
  double s21;
  double s22;
  double z;
  bool gen;
} RNDState;

class OpenCLBase {

private:
	
  static cl_platform_id m_platform_id;
  static cl_device_id m_device_id;

  cl_context_properties m_context_properties[3];
  cl_program m_program;
  cl_kernel m_kernel;
	
  static cl_event m_last_event;
  cl_int m_num_events;
  std::vector<cl_event> m_events;
	
  char * m_kernel_file;

  cl_device_type m_device_type;
	
  /*
    Name: getPlatforms
    Info: get all avaialble platforms and save in m_platform_ids, save number of platforms
    Return: success or error code
  */
  int ocl_getPlatforms();
	
	
  /*
    Name: getDevice
    Info: get first avaialble devices and save device id and platform id for this device, device name: (-gpu, -mic, -cpu)
    ReturnL success or error code
  */
  int ocl_getDevice(const char* device_name);
	
  /*
    Name getDeviceType
    Info: get device type from device name (-gpu, -cpu, -mic)
    Return: success or error code
  */
  int ocl_getDeviceType(const char* device_name, cl_device_type &device_type);
	
  /*
    Name: createContext
    Info: create context with specified device
    Return: success or error code
  */
  int ocl_createContext();
	
  /*
    Name: buildProgram
    Info: build program from specified kernel file
    Return: success or error code
  */
  int ocl_buildProgram(const char* kernel_file);

  /** Compile program from kernel source string
   *
   */
  int ocl_compileProgram(const char* kernel_source, const char* opts = NULL);

protected:

  int defaultRndSet;
  cl_mem defaultRndState;
	
	
public:

  static cl_context m_context;
  static cl_command_queue m_command_queue;
    
  /*
    constructor
  */
  OpenCLBase();
    
  /*
    destructor
  */
  ~OpenCLBase();
    
  /*
    Create RND states
    Return: success or error code
  */
  int ocl_createRndStates(int size);

  /* Create an array of random numbers on the device
   *
   */
  int ocl_createRandomNumbers(void *mem_ptr, int size);

  /*
    Destroy rnd states
    Return: success or error code
  */
  int ocl_deleteRndStates();


  /*
    Name: getAllDevices
    Info: get all available devices
    ReturnL success or error code
  */
  int ocl_getAllDevices();

  /** Get the OpenCL device count for the set type of device
   *
   */
  int ocl_getDeviceCount(int &ndev);

  /** Get the name of the device used
   */
  int ocl_getDeviceName(std::string &device_name);

  /** Set the device to use for OpenCL kernels.
   *  device id to use is passed as integer.
   */
  int ocl_setDevice(int device);

  /** Get a list of all the unique devices of the same type that can run OpenCL kernels
   *  Used when GPUs of different types might be pressent on the system.
   */
  int ocl_getUniqueDevices(std::vector<int> &devices);
    
  /*
    Name: setUp
    Info: set up opencl resources
    Return: success or error code
  */
  int ocl_setUp(const char* device_name);
	
  /*
    Name: loadKernel
    Info: load and compile opencl kernel file if it has changed
    Return: success or error code
  */
  int ocl_loadKernel(const char* kernel_file);


  /** Build program from kernel source.
   * Builds a program from source code provided in kernel_source.
   * If compilation fails will return DKS_ERROR
   */
  int ocl_loadKernelFromSource(const char* kernel_source, const char* opts = NULL);
	
  /*
    Name: allocateMemory
    Info: allocate memory on device
    Return: return pointer to memory
  */
  cl_mem ocl_allocateMemory(size_t size, int &ierr);

  /*
    Name: allocateMemory
    Info: allocate memory on device
    Return: return pointer to memory
  */
  cl_mem ocl_allocateMemory(size_t size, int type, int &ierr);
	
  /** Zero OpenCL memory buffer
   *  Set all the elemetns in the device array to zero
   */
  template <typename T>
  int ocl_fillMemory(cl_mem mem_ptr, size_t size, T value, int offset = 0) {
  
    cl_int ierr;
    ierr = clEnqueueFillBuffer(m_command_queue, mem_ptr, &value, sizeof(T), offset, 
			       sizeof(T)*size, 0, nullptr, nullptr);
    if (ierr != CL_SUCCESS)
      return DKS_ERROR;
    return DKS_SUCCESS;
  }

  /*
    Name: writeData
    Info: write data to device memory (needs ptr to mem object)
    Return: success or error code
  */
  int ocl_writeData(cl_mem mem_ptr, const void * in_data, size_t size, size_t offset = 0, int blocking = CL_TRUE);
	
  /*
    Name: copyData
    Info: copy data from one buffer on the device to another
    Return: success or error code
  */
  int ocl_copyData(cl_mem src_ptr, cl_mem dst_ptr, size_t size);
	
  /*
    Name: createKernel
    Info: create kernel from program
    Return: success or error code
  */
  int ocl_createKernel(const char* kernel_name);
	
  /*
    Name: setKernelArgs
    Info: set opencl kernel arguments
    Return: success or error code
  */
  int ocl_setKernelArg(int idx, size_t size, const void *arg_value);
	
  /*
    Name: executeKernel
    Info: execute selected kernel (needs kernel parameters)
    Return: success or error code
  */
  int ocl_executeKernel(cl_uint, const size_t *work_items, const size_t *work_grou_size = NULL);
	
  /*
    Name: readData
    Info: read data from device (needs pointer to mem object)
    Return: success or error code
  */
  int ocl_readData(cl_mem mem_ptr, void * out_data, size_t size, size_t offset = 0, int blocking = CL_TRUE);
	
  /*
    Name: freeMemory
    Info: free device memory (needs ptr to mem object)
    Return: success or error code
  */
  int ocl_freeMemory(cl_mem mem_ptr);
	
  /*
    Name: cleanUp
    Info: free opencl resources
    Return: success or error code
  */
  int ocl_cleanUp();
	
  /*
    Name: deviceInfo
    Info: print device info (mostly for debugging purposes)
    Return: success or error code
  */
  int ocl_deviceInfo(bool verbose = true);

  /* Check OpenCL kernel.
   * Query device and check if it can run the kernel with required parameters
   */
  int ocl_checkKernel(const char* kernel_name, int work_group_size,
		      bool double_precision, int &threadsPerBlock);
	
  /*
    Name: clearEvents
    Info: clear saved events (for debuging purposes)
    Return: nothing
  */
  void ocl_clearEvents();

  /*
    Name: eventInfo
    Info: print information about kernel timings (for debuging purposes)
    Return: nothing
  */
  void ocl_eventInfo();

  /*
    Return current command queue
  */
  cl_command_queue ocl_getQueue() { return m_command_queue; }
};

#endif








